/*
 * OSD_flash.c - OSD-Flash chip driver over SPI port
 *
 * Copyright 2017 Sony Corporation
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/uaccess.h>    /* copy_from_user */

#include <linux/spi/spi.h>
#include <linux/delay.h>      /* msleep() */
#include <linux/poll.h>       /* .poll */
#include <linux/wait.h>       /* wait_queue_head_t */

#include "../include/OSD_flash.h"


struct flash_chr_dev_t {
	dev_t id;
	struct cdev cdev;
	struct class *class;
	struct device *dev;
} *flash_chr_dev;

#define ADDR_OFFSET_MASK    0xFF
#define FLASH_SPI_FIFO_SIZE FLASH_RW_DATA_SIZE

/*
 * This supports access to SPI devices using normal userspace I/O calls.
 * Note that while traditional UNIX/POSIX I/O semantics are half duplex,
 * and often mask message boundaries, full SPI support requires full duplex
 * transfers.  There are several kinds of internal message boundaries to
 * handle chipselect management and other protocol options.
 *
 * SPI has a character major number assigned.  We allocate minor numbers
 * dynamically using a bitmask.  You must use hotplug tools, such as udev
 * (or mdev with busybox) to create and destroy the /dev/spidevB.C device
 * nodes, since there is no fixed association of minor numbers with any
 * particular SPI bus or device.
 */
#define N_SPI_MINORS      32  /* ... up to 256 */

static DECLARE_BITMAP(minors, N_SPI_MINORS);

/* expected to link to spi driver */
struct spidev_data {
	dev_t             devt;
	struct spi_device *spi;
	wait_queue_head_t qh_notify_all;
};

/* keep interrupt status registers for notifying to each processes */
struct notify_registers_list {
	struct list_head list;
};

/* notify information for each processes */
/* they can be linked with filep's */
struct intr_notify_list {
	struct file *filp;
	struct spidev_data *spidev;
	int notify_list_inited;
	struct notify_registers_list notify_registers_list_head;
	struct list_head list;
};
struct intr_notify_list intr_notify_list_head;

static struct spidev_data *g_spidev;

static uint8_t  *rw_buffer;	/* Read Write Buffer */
static uint32_t WriteAddress;
static uint32_t ReadAddress;

static void addr_convert( const uint32_t addr, uint8_t* const cmd )
{
	/* cmd[0] is actual command */
	cmd[1] = (addr >> 16) & ADDR_OFFSET_MASK;
	cmd[2] = (addr >>  8) & ADDR_OFFSET_MASK;
	cmd[3] =  addr        & ADDR_OFFSET_MASK;
}

static int flash_send_command( const struct spi_device* const spi,
							   const uint8_t  cmd,
							   const uint8_t* const wr_ptr,
							         uint8_t* const rd_ptr,
							   const uint8_t  len)
{
	struct spi_transfer xfer;
	uint8_t tx_buffer[FLASH_RW_DATA_SIZE];
	uint8_t rx_buffer[FLASH_RW_DATA_SIZE];
	int ret;
	struct spi_message msg;

//  pr_info(": flash_send_command()\n");

	if( spi == NULL ) {
		return FLASH_DRV_ERR;
	}

	if( len > FLASH_RW_DATA_SIZE ) {
		return FLASH_DRV_ERR;
	}

	spi_message_init( &msg );

	memset( tx_buffer, 0, sizeof( tx_buffer ) );
	memset( rx_buffer, 0, sizeof( rx_buffer ) );

	tx_buffer[0] = cmd;

	xfer.tx_buf = tx_buffer;
	xfer.rx_buf = rx_buffer;

	if( cmd == FLASH_NORMAL_READ_DATA ) {

		tx_buffer[1] = wr_ptr[0];
		tx_buffer[2] = wr_ptr[1];
		tx_buffer[3] = wr_ptr[2];

		xfer.len = FLASH_RW_DATA_SIZE;

	} else {
		xfer.len = len;

		if( wr_ptr != NULL ) {
			tx_buffer[1] = *wr_ptr;
		}
	}

	spi_message_add_tail( &xfer, &msg );
	ret = spi_sync( (struct spi_device*)spi, &msg );
	if( ret < 0 ) {
		dev_warn(&spi->dev, "Message transfer err:%d\n", ret);
		return FLASH_DRV_NG;
	}

	if( cmd == FLASH_NORMAL_READ_DATA ) {
		memcpy( rd_ptr, &rx_buffer[FLASH_RW_CMD_OFFSET_LEN], FLASH_RW_DATA_LEN );
	} else {
		if( rd_ptr != NULL ) {
			memcpy( rd_ptr, &rx_buffer[1], len );
		}
	}

	return FLASH_DRV_OK;
}

static int flash_send_command_multi( const struct spi_device* const spi,
							   const uint8_t* const cmd_offset,
							   const uint8_t* const wr_ptr,
							         uint8_t* const rd_ptr,
							   const uint8_t  len)
{
	struct spi_transfer xfer;
	uint8_t tx_buffer[FLASH_RW_DATA_SIZE];
	uint8_t rx_buffer[FLASH_RW_DATA_SIZE];
	int ret;
	struct spi_message msg;

//  pr_info(": flash_send_command_multi()\n");

	if( spi == NULL ) {
		return FLASH_DRV_ERR;
	}

	if( cmd_offset == NULL ) {
		return FLASH_DRV_ERR;
	}

	if( len > FLASH_RW_DATA_LEN ) {
		return FLASH_DRV_ERR;
	}

	spi_message_init(&msg);


	xfer.tx_buf = tx_buffer;
	xfer.rx_buf = rx_buffer;
	xfer.len    = len + FLASH_RW_CMD_OFFSET_LEN;

	memset( tx_buffer,          0, sizeof(tx_buffer) );
	memcpy( tx_buffer, cmd_offset, FLASH_RW_CMD_OFFSET_LEN );

	if( wr_ptr != NULL ) {
		memcpy( &tx_buffer[FLASH_RW_CMD_OFFSET_LEN], wr_ptr, len );
	}

	spi_message_add_tail( &xfer, &msg );
	ret = spi_sync( (struct spi_device*)spi, &msg );
	if( ret < 0 ) {
		dev_warn( &spi->dev, "Message transfer err:%d\n", ret );
		return FLASH_DRV_NG;
	} else {
		if( rd_ptr != NULL ) {
		    memcpy( rd_ptr, &rx_buffer[FLASH_RW_CMD_OFFSET_LEN], len );
		}
	}

	return FLASH_DRV_OK;

}

static int flash_cmd_wait_ready( const struct spi_device* const spi, const uint32_t timeout )
{
	int ret;
	uint8_t status;
	uint32_t timebase = timeout;

	if( spi == NULL ) {
		return FLASH_DRV_ERR;
	}

	do {

		/* Read Status */
		ret = flash_send_command( spi, FLASH_READ_STATUS1, NULL, &status, 2 );
		if( ret ) {
			break;
		}
		if( (status & STATUS_WIP) == 0 ) {
			ret = FLASH_DRV_OK;
			break;
		}
		udelay( WAIT_READY_TIME );

	} while( timebase-- );

	if( timebase == 0 ) {
		/* Timed out */
		pr_err( "OSD_flash.ko: Ready Time Out\n" );
		ret = FLASH_DRV_NG;
	}

	return ret;
}

static int OSD_flash_erase( const struct spi_device* const spi, const uint8_t cmd )
{
	int ret = FLASH_DRV_OK;

	pr_info( ": OSD_flash_erase()\n" );

	if( spi == NULL ) {
		return FLASH_DRV_ERR;
	}

	if( cmd != FLASH_CHIP_ERASE ) {
		return FLASH_DRV_ERR;
	}

	do {

		/* Write Enable Command */
		ret = flash_send_command( spi, FLASH_SET_WRITE_ENABLE, NULL, NULL, 1 );
		if( ret ) {
			break;
		}

		udelay( 1 );

		/* Erase Command */
		ret = flash_send_command( spi, cmd, NULL, NULL, 1 );
		if( ret ) {
			break;
		}

		/* Wait */
		ret = flash_cmd_wait_ready( spi, CHIP_ERASE_TIMEOUT );
		if( ret ) {
			break;
		}

	} while( 0 );

	return ret;
}

/*
 * OSD_flash driver file operations
 */

static int OSD_flash_open( struct inode *inode, struct file *filp )
{
	struct intr_notify_list *notify_list;
	int iRet = FLASH_DRV_OK;

	pr_info(": OSD_flash_open()\n");

	do {

		if( filp == NULL ) {
			iRet = FLASH_DRV_ERR;
			break;
		}

		if( g_spidev == NULL ) {
			pr_err( "OSD_flash.ko: open failed - device driver is not ready\n" );
			iRet = -ENODEV;
			break;
		}

		/* keep spi driver information in filp */
		notify_list = kzalloc( sizeof(*notify_list), GFP_KERNEL );
		if ( notify_list == NULL ) {
			iRet = -ENOMEM;
			break;
		}

		/* Read Write Buffer */
		rw_buffer = kmalloc( FLASH_RW_DATA_SIZE, GFP_KERNEL );
		if ( rw_buffer == NULL ) {
			pr_err( "OSD_flash.ko: open failed - device driver is not ready\n" );
			iRet = -ENOMEM;
			kfree( notify_list );
			break;
		}

		/* Read Write Address Initial */
		WriteAddress = 0;
		ReadAddress  = 0;

		notify_list->spidev = g_spidev;
		notify_list->filp   = filp;
		filp->private_data  = notify_list;
		nonseekable_open( inode, filp );

	} while( 0 );

	return iRet;
}

static int OSD_flash_close( struct inode *inode, struct file *filp )
{
	struct intr_notify_list *notify_list;
	struct spidev_data *spidev;

	pr_info(": OSD_flash_close()\n");

	notify_list = (struct intr_notify_list *)filp->private_data;
	spidev = notify_list->spidev;

	if( notify_list->notify_list_inited ) {

		int hit = 0;
		struct intr_notify_list *st_notify;
		struct list_head *list;

		list_for_each( list, &intr_notify_list_head.list ) {
			st_notify = list_entry( list, struct intr_notify_list, list );
			dev_dbg( &spidev->spi->dev, "search filp=0x%p   delete 0x%p\n", st_notify->filp, filp );
			if( st_notify->filp == filp ) {
				hit = 1;
				break;
			}
		}
		if( hit ) {
			dev_dbg( &spidev->spi->dev, "removed filp=0x%p\n", st_notify->filp );
			list_del( list );
			/* struct intr_norify_list = filp->private_data is freed below */
		} else {
			dev_err( &spidev->spi->dev, "could not find notify_list to remove\n" );
		}

		/* remove all interrupt status register data, if they are in the list */
		{
			struct notify_registers_list *st_regs;
			struct list_head *list;

			list_for_each( list, &notify_list->notify_registers_list_head.list ) {
				st_regs = list_entry( list, struct notify_registers_list, list );
				list_del( list );
				kfree( st_regs );
			}
	    }
	}

	kfree( rw_buffer );	/* Read Write Buffer */

	kfree( filp->private_data ); /* struct intr_notify_list */
	filp->private_data = NULL;

	return FLASH_DRV_OK;
}

static ssize_t OSD_flash_read( struct file *filp, char __user *buf, size_t size, loff_t *offset )
{
	struct intr_notify_list *notify_list;
	struct spidev_data  *spidev;
	struct spi_device *spi;
	uint8_t cmd[4];
	int iRet = 0;

//	pr_info(": OSD_flash_read()\n");

	if( size > (FLASH_SPI_FIFO_SIZE - FLASH_RW_CMD_OFFSET_LEN) ) {
		pr_err("OSD_flash.ko: xfer len is invalid over fifo size - \n");
		return FLASH_DRV_ERR;
	}

	notify_list = (struct intr_notify_list *)filp->private_data;
	spidev = notify_list->spidev;
	spi = spi_dev_get( spidev->spi );

	if( spi == NULL ) {
		return -ESHUTDOWN;
	}

	if( ReadAddress > 0x00FFFFF0 ) {
		return FLASH_DRV_NG;
	}

	memset( rw_buffer, 0, FLASH_RW_DATA_SIZE );

	cmd[0] = FLASH_NORMAL_READ_DATA;
	addr_convert( ReadAddress, cmd );

	/* Data Read */
	iRet = flash_send_command_multi( spi, cmd, NULL, rw_buffer, size );
	if( iRet < 0 ) {
		pr_err("OSD_flash.ko: Read Data failed - \n");
		return iRet;
	}
	iRet = copy_to_user(buf, rw_buffer, size);
	if( iRet ) {
		return -EFAULT;
	}

	/* Offset Address Update */
	ReadAddress += size;

	return FLASH_DRV_OK;
}

static ssize_t OSD_flash_write( struct file *filp, const char __user *buf, size_t size, loff_t *offset )
{
	struct intr_notify_list *notify_list;
	struct spidev_data  *spidev;
	struct spi_device *spi;
	int ret;
	uint8_t cmd[4];

	if( size > (FLASH_SPI_FIFO_SIZE - FLASH_RW_CMD_OFFSET_LEN) ) {
		pr_err( "OSD_flash.ko: xfer len is invalid over fifo size - \n" );
		return FLASH_DRV_ERR;
	}

	notify_list = (struct intr_notify_list *)filp->private_data;
	spidev = notify_list->spidev;
	spi = spi_dev_get( spidev->spi );

	if( spi == NULL ) {
		return -ESHUTDOWN;
	}

	memset( rw_buffer, 0, FLASH_RW_DATA_SIZE );
	ret = copy_from_user( rw_buffer, buf, size );
	if( ret ) {
		return -EFAULT;
	}

	if( WriteAddress > 0x00FFFFF0 ) {
		return FLASH_DRV_NG;
	}

	cmd[0] = FLASH_SET_PAGE_PROGRAM;
	addr_convert( WriteAddress, cmd );

	/* Write Enable Command */
	ret = flash_send_command( spi, FLASH_SET_WRITE_ENABLE, NULL, NULL, 1 );
	if( ret ) {
		pr_err( "OSD_flash.ko: Write Enable failed - \n" );
		return FLASH_DRV_NG;
	}

	udelay( 1 );

	/* Data Write */
	ret = flash_send_command_multi( spi, cmd, rw_buffer, NULL, size );
	if( ret < 0 ) {
		pr_err( "OSD_flash.ko: Page Program failed - \n" );
		return FLASH_DRV_NG;
	}

	/* Wait */
	ret = flash_cmd_wait_ready( spi, PAGE_PROGRAM_TIMEOUT );
	if( ret ) {
		pr_err( "OSD_flash.ko: Write Time Out - \n" );
		return FLASH_DRV_NG;
	}

	/* Write Disable Command */
	ret = flash_send_command( spi, FLASH_SET_WRITE_DISABLE, NULL, NULL, 1 );
	if( ret ) {
		pr_err( "OSD_flash.ko: Write Enable failed - \n" );
		return FLASH_DRV_NG;
	}

	udelay( 1 );

	/* Offset Address Update */
	WriteAddress += size;

	return FLASH_DRV_OK;
}

static long OSD_flash_ioctl( struct file *filp, unsigned int cmd, unsigned long arg )
{
	struct intr_notify_list *notify_list;
	struct spidev_data  *spidev;
	int     err = 0;
	int     retval = 0;
	struct spi_device *spi;
	struct flash_dev_ioctl_params param;
	uint8_t rd_data[FLASH_RW_DATA_SIZE], wr_data[FLASH_RW_DATA_SIZE];

//  pr_info(": OSD_flash_ioctl()\n");

	/* guard against device removal before, or while,
	 * we issue this ioctl.
	*/

	notify_list = (struct intr_notify_list *)filp->private_data;
	spidev = notify_list->spidev;
	spi = spi_dev_get( spidev->spi );

	if( spi == NULL ) {
	  return -ESHUTDOWN;
	}

	memset( &param, 0, sizeof(struct flash_dev_ioctl_params) );
	err = copy_from_user( &param, (const void *)arg, FLASH_IOCTL_MAX_LEN );
	if( err ) {
		/* some of the data were not able to copy */
		return -EFAULT;
	}

	/* use the buffer lock here for triple duty:
	 *  - prevent I/O (from us) so calling spi_setup() is safe;
	 *  - prevent concurrent SPI_IOC_WR_* from morphing
	 *    data fields while SPI_IOC_RD_* reads them;
	 *  - SPI_IOC_MESSAGE needs the buffer locked "normally".
	 */

//	pr_info(": OSD_flash_ioctl() cmd = 0x%02X\n", param.cmd);
	memset( rd_data, 0, sizeof( rd_data ) );
	memset( wr_data, 0, sizeof( wr_data ) );

	switch( param.cmd ) {

		case FLASH_READ_JEDEC_ID:
			retval = flash_send_command( spi, param.cmd, NULL, rd_data, 4 );
			if( retval ) {
				break;
			}

			param.lvalue  = 0;
			param.lvalue |= (rd_data[0] << 16) & 0x00FF0000;
			param.lvalue |= (rd_data[1] <<  8) & 0x0000FF00;
			param.lvalue |=  rd_data[2]        & 0x000000FF;

			err = copy_to_user( (void *)arg, &param, sizeof( struct flash_dev_ioctl_params ) );
			if( err ) {
				retval = -EFAULT;
			}
			break;

		case FLASH_READ_STATUS1:
		case FLASH_READ_STATUS2:
		case FLASH_READ_STATUS3:

			retval = flash_send_command( spi, param.cmd, NULL, rd_data, 2 );
			if( retval ) {
				break;
			}

			param.value = rd_data[0];
			dev_dbg( &spi->dev, "ioctl jedec ID read  result 0x%X\n", param.value );
			err = copy_to_user( (void *)arg, &param, sizeof( struct flash_dev_ioctl_params) );
			if( err ) {
				retval = -EFAULT;
			}
			break;

		case FLASH_WRITE_STATUS1:
		case FLASH_WRITE_STATUS2:
		case FLASH_WRITE_STATUS3:

			wr_data[0] = param.value;
			retval = flash_send_command( spi, param.cmd, wr_data, rd_data, 2 );
			break;

		case FLASH_SET_WRITE_ENABLE:
		case FLASH_SET_SR_WRITE_ENABLE:
		case FLASH_SET_ENABLE_RESET:

			retval = flash_send_command( spi, param.cmd, NULL, rd_data, 1 );
			break;

		case FLASH_SECTOR_ERASE:
			break;

		case FLASH_BLOCK32_ERASE:
		case FLASH_BLOCK64_ERASE:

			retval = flash_send_command( spi, param.cmd, wr_data, NULL, 1 );
			break;

		case FLASH_CHIP_ERASE:

			retval = OSD_flash_erase( spi, param.cmd );
			break;

		case FLASH_NORMAL_READ_DATA:

			wr_data[0] = (param.lvalue >> 16) & ADDR_OFFSET_MASK;
			wr_data[1] = (param.lvalue >>  8) & ADDR_OFFSET_MASK;
			wr_data[2] =  param.lvalue        & ADDR_OFFSET_MASK;

			retval = flash_send_command( spi, param.cmd, wr_data, rd_data, FLASH_RW_DATA_LEN );
			break;

			default:
				retval = -EINVAL;
				break;
	}

	spi_dev_put( spi );
	return retval;
}

/* to link with spi driver */
static int OSD_flash_probe( struct spi_device* const spi )
{
	struct spidev_data *spidev;
	unsigned long      minor;

	pr_info(": OSD_flash_probe()\n");

	if( spi == NULL ) {
	  return -ESHUTDOWN;
	}

	/* Allocate driver data */
	spidev = kzalloc( sizeof( *spidev ), GFP_KERNEL );
	if( spidev == NULL ) {
		return -ENOMEM;
	}

	/* Initialize the driver data */
	spidev->spi = spi;

	minor = find_first_zero_bit( minors, N_SPI_MINORS );

	dev_dbg( &spi->dev, "before create_device   OSD_flash%d.%d\n", spi->master->bus_num, spi->chip_select );

	if( minor < N_SPI_MINORS ) {
		spidev->devt = MKDEV( MAJOR( flash_chr_dev->id ), minor );
		set_bit( minor, minors );
		spi_set_drvdata( spi, spidev );
		g_spidev = spidev;
	} else {
		spi_set_drvdata( spi, NULL );
		kfree( spidev );
		g_spidev = NULL;
		dev_err( &spi->dev, "OSD_flash_spi_probe failed %d\n", -ENODEV );
		return -ENODEV;
	}

	INIT_LIST_HEAD( &intr_notify_list_head.list );
	init_waitqueue_head( &spidev->qh_notify_all );

	return FLASH_DRV_OK;
}

struct spi_device_id OSD_flash_id_table = {"OSD_flash_spi", 0};

const static struct spi_driver flash_driver_st = {
	.driver = {
		.name  = FLASH_DRV_DEVNAME,
		.bus   = &spi_bus_type,
		.owner = THIS_MODULE,
	},
	.probe    = OSD_flash_probe,
	.id_table = &OSD_flash_id_table,
};

const struct file_operations OSD_flash_fops = {
	.owner          = THIS_MODULE,
	.open           = OSD_flash_open,
	.release        = OSD_flash_close,
	.read           = OSD_flash_read,
	.write          = OSD_flash_write,
	.unlocked_ioctl = OSD_flash_ioctl,
};

static int __init OSD_flash_init( void )
{
	int ret;

	pr_info( ": OSD_flash_init()\n" );

	do {

		flash_chr_dev = kmalloc( sizeof( struct flash_chr_dev_t ), GFP_KERNEL );
		if( flash_chr_dev == NULL ) {
			pr_err("kmalloc err\n");
			ret = -ENOMEM;
			break;
		}

		ret = alloc_chrdev_region( &flash_chr_dev->id, 0, 1, FLASH_DRV_DEVNAME );
		if( ret ) {
			pr_err( "alloc_chrdev_region err %d\n", ret );
			break;
		}

		cdev_init( &flash_chr_dev->cdev, &OSD_flash_fops );
		flash_chr_dev->cdev.owner = THIS_MODULE;

		ret = cdev_add( &flash_chr_dev->cdev, flash_chr_dev->id, 1 );
		if( ret ) {
			pr_err( "cdev_add err %d\n", ret );
			break;
		}

		flash_chr_dev->class = class_create( THIS_MODULE, FLASH_DRV_DEVNAME );
		if( IS_ERR(flash_chr_dev->class) ) {
			ret = PTR_ERR( flash_chr_dev->class );
			pr_err( "class_create err %d\n", ret );
			break;
		}

		flash_chr_dev->dev = device_create( flash_chr_dev->class, NULL, flash_chr_dev->id, NULL, FLASH_DRV_DEVNAME );
		if( IS_ERR(flash_chr_dev->dev) ) {
			ret = PTR_ERR( flash_chr_dev->dev );
			pr_err( "device_create err %d\n", ret );
			break;
		}

		/* spi_register_board_info() is called from model_init() */
		ret = spi_register_driver( (struct spi_driver*)&flash_driver_st );
		if( ret < 0 ) {
			pr_err( "spi_device_register(OSD_flash.c) err %d\n", ret );
			break;
		} else {
			ret = 0;
		}

	} while( 0 );

	if( ret ) {
		class_destroy( flash_chr_dev->class );
		cdev_del( &flash_chr_dev->cdev );
		unregister_chrdev_region( flash_chr_dev->id, 1 );
		kfree( flash_chr_dev );
	}
	pr_info( ": OSD_flash_init() End %d\n", ret ); /* dev_info() is not available. */

	return ret;
}

static void __exit OSD_flash_exit( void )
{
	pr_info( "spidev:exit\n" );

	spi_unregister_driver( (struct spi_driver*)&flash_driver_st );

	device_destroy( flash_chr_dev->class, flash_chr_dev->id );
	class_destroy( flash_chr_dev->class );
	cdev_del( &flash_chr_dev->cdev );
	unregister_chrdev_region( flash_chr_dev->id, 1 );
	kfree( flash_chr_dev );
}

module_init( OSD_flash_init );
module_exit( OSD_flash_exit );

MODULE_LICENSE( "GPL" );
MODULE_DESCRIPTION( "Flash driver" );
